@@ -1,7 +1,7 @@
 /* x86_features.h -- check for CPU features
-* Copyright (C) 2013 Intel Corporation Jim Kukunas
-* For conditions of distribution and use, see copyright notice in zlib.h
-*/
+ * Copyright (C) 2013 Intel Corporation Jim Kukunas
+ * For conditions of distribution and use, see copyright notice in zlib.h
+ */
 
 #ifndef X86_FEATURES_H_
 #define X86_FEATURES_H_
@@ -14,8 +14,10 @@ struct x86_cpu_features {
     int has_avx512vl;
     int has_avx512_common; // Enabled when AVX512(F,DQ,BW,VL) are all enabled.
     int has_avx512vnni;
+    int has_bmi2;
     int has_sse2;
     int has_ssse3;
+    int has_sse41;
     int has_sse42;
     int has_pclmulqdq;
     int has_vpclmulqdq;
@@ -25,4 +27,4 @@ struct x86_cpu_features {
 
 void Z_INTERNAL x86_check_features(struct x86_cpu_features *features);
 
-#endif /* CPU_H_ */
+#endif /* X86_FEATURES_H_ */
